import sys

class PosReader:
    def __init__(self, f, tb):
        self.f = f
        self.tb = tb
    def __iter__(self):
        for line in self.f:
            if line.startswith('#'): 
                continue
            L = line.split()
            chrom, pos, ref, alt = L[:4]
            if not (len(ref) == 1 and len(alt) == 1):
                print("Warning. Non-SNV variant ignored:", line.strip(), file=sys.stderr)
                continue
            if ref != self.tb.sequence(chrom, int(pos)-1, int(pos)):
                print(int(pos))
                print("Warning. Reference allele dosn't match 2bit file:", line.strip(), "ref:", self.tb.sequence(chrom, int(pos)-1, int(pos)), file=sys.stderr)
                continue
            yield chrom, int(pos)-1, 1

class BedReader():
    def __init__(self,f):
        self.f = f
    def __iter__(self):
        for line in self.f:
            L = line.split()
            chrom, start, end = L[:3]
            for pos in range(int(start), int(end)):
                yield chrom, pos, 1

class AllAutoReader():
    def __init__(self, tb):
        self.tb = tb
        if any(x.startswith('chr') for x in tb.chroms()):
            prefix = 'chr'
        else:
            prefix = ''
        self.autosomes = [prefix + str(x) for x in range(1,23)]
    def __iter__(self):
        for chrom in self.autosomes:
            start = 0
            end = self.tb.chroms()[chrom]
            for pos in range(int(start), int(end)):
                yield chrom, pos, 1


"""
kmer_counter package.

Count kmers in regions or at SNVs or at indel breakpoints.
"""
from kmer_counter.utils import *

__all__: [reverse_complement, get_indel_contexts, get_possible_indel_pos]

__version__ = "0.2.2"



import sys
complement = str.maketrans('ATCGN', 'TAGCN')

def reverse_complement(s):
    """Reverse complement

    Args:
        s (str): input string should contain only A,C,G or T 

    Returns:
        str: reverse complement of input
    """
    return s.translate(complement)[::-1]

def get_indel_contexts(chrom, pos, radius, tb):
    """get sequence context around an indel breakpoint

    Args:
        chrom (str): chromosome
        pos (int): position
        radius (int): flank 
        tb: TwoBit object

    Raises:
        Exception: If context contains 'N' or length is smaller than expected

    Returns:
        (str,str): context and reverse complement of context.
    """
    context = tb.sequence(chrom, pos-radius, pos+radius)
    if 'N' in context or len(context) < radius*2:
        raise Exception('bad_context')
    context2 = reverse_complement(context)
    return (context, context2)


def get_possible_indel_pos(chrom, pos, ref, alt, tb, buffer=50):
    """Get list with possible positions of an indel
    This method assumes that the indel is left aligned.
    But does not check if that is true.

    Args:
        chrom (str): chromosome
        pos (int): position
        ref (str): reference base
        alt (str): alternative base
        tb (): TwoBit object
        buffer (int, optional): Number of upstream bases to read. Defaults to 50.
            If it turns out to be to small it will automatically be increased.

    Returns:
        list of int: all possible positions of the indel
    """

    if buffer <= len(alt)+len(ref):
        return get_possible_indel_pos(chrom, pos, ref, alt, tb, buffer*2)
    
    # long_seq og short_seq starts at the first position not shared between
    # alt og ref
    if len(ref) > len(alt): # This variant is a deletion
        assert(len(alt)==1)
        assert(ref[0] == alt)
        #print(chrom, pos, ref, alt)        
        long_seq = tb.sequence(chrom, pos, pos+buffer) # Reference sequence
        #print(long_seq)
        #print(long_seq[:len(ref)-1],ref[1:])
        #print(tb.sequence(chrom, pos-1, pos))
        if(long_seq[:len(ref)-1] != ref[1:]):
            print(f"WARNING. Reference does not match: {chrom} {pos} {ref} {alt}", file=sys.stderr)
            print(f"ref according to mutations file: {ref}", file=sys.stderr)
            oref = tb.sequence(chrom, pos-1, pos+len(ref)-1)
            print(f"ref according to 2bit file: {oref}", file=sys.stderr)
            print(f"Skipping variant.", file=sys.stderr)
            return None
        #assert(long_seq[:len(ref)-1] == ref[1:])
        short_seq = long_seq[len(ref)-1:] # Alternative sequence
        i = 0
    elif len(ref) < len(alt): # This variant is an Insertion
        assert(len(ref)==1)
        assert(ref == alt[0])
        short_seq = tb.sequence(chrom, pos, pos+buffer) # Reference seqeunce
        long_seq =  alt[1:] + short_seq # Alternative seqeunce
        i = 0 
    else:
        assert(False)
    L = []
    len_diff = len(long_seq) - len(short_seq)
    assert(short_seq[:i] == long_seq[:i] and short_seq[i:] == long_seq[i+len_diff:])
    while True:     
        #print("prefixes = ", short_seq[:i], long_seq[:i])
        #print("suffix = ", short_seq[i:], long_seq[i+len_diff:])
        if short_seq[:i] == long_seq[:i] and short_seq[i:] == long_seq[i+len_diff:]:
            L.append((i+pos, long_seq[i:i+len_diff]))            
            i+=1
            if i + len_diff >= buffer:
                return get_possible_indel_pos(chrom, pos, ref, alt, tb, buffer*2)
        else:
            return L


import random
import sys
from collections import defaultdict
from kmer_counter.utils import *

def count_indels(mutations, tb, break_point_type, radius, sample_breakpoints=True):
    """ Count k-mers at indels breakpoints

    Args:
        mutations: file with mutations
        tb: TwoBit file.
        break_point_type (str): type of breakpoints that should be counted
        radius (int): number of upstream and downstream bases to consider.
        sample_breakpoints (bool, optional): Should breakpoints be sampled? Defaults to True.
    Returns:
        dict: kmers -> counts
            if sample_breakpoints==True then counts will be ints
            else counts will be floats.
    """    
    if sample_breakpoints:
        kmer_count = defaultdict(int)
    else:        
        kmer_count = defaultdict(float)

    for line in mutations:
        chrom, pos, ref, alt = line.split()[:4]
        #if args.verbose:
        #    print(chrom, pos, ref, alt)#, file=sys.stderr)
        pos = int(pos)
        if len(ref) > len(alt):
            if len(alt) != 1:
                if (ref[-(len(alt)-1):] != alt[1:]):
                    print("Warning. Complex variant ignored:", line.strip(), file=sys.stderr)
                    continue
                #print("Variant before", alt, ref)
                ref = ref[:-(len(alt)-1)]
                alt = alt[:1]
                #print("Variant after", alt, ref)
        elif len(alt)>len(ref):
            if len(ref) != 1:
                if(ref[1:] != alt[-(len(ref)-1):]):
                    print("Warning. Complex variant ignored:", line.strip(), file=sys.stderr)
                    continue
                alt = alt[:-(len(ref)-1)]
                ref = ref[:1]
        if alt[0] != ref[0]:
            print("Warning. Not left-aligned variant ignored:", line.strip(), file=sys.stderr)
            continue
        L = get_possible_indel_pos(chrom, pos, ref, alt, tb)
        if L is None:
            continue
        if sample_breakpoints:
            L = [random.choice(L)]
        for rpos, ralt in L:
            if len(ref) < len(alt): # This variant is an Insertion
                try:
                    c1, c2 = get_indel_contexts(chrom, rpos, radius, tb)
                except:
                    continue
                if break_point_type in ['ins', 'all']:
                    if sample_breakpoints:
                        kmer_count[c1] += 1
                        kmer_count[c2] += 1
                    else:
                        kmer_count[c1] += 1.0/(len(L)*2)
                        kmer_count[c2] += 1.0/(len(L)*2)
            elif len(ref) > len(alt): # This variant is a Deletion
                try:
                    c1, c2 = get_indel_contexts(chrom, rpos, radius, tb)
                except:
                    continue
                if break_point_type in ['del_start', 'all', 'del']:
                    if sample_breakpoints:
                        kmer_count[c1] += 1
                    else:
                        kmer_count[c1] += 1.0/(len(L)*4)
                if break_point_type in ['del_end', 'all', 'del']:
                    if sample_breakpoints:
                        kmer_count[c2] += 1
                    else:
                        kmer_count[c2] += 1.0/(len(L)*4)
                try:
                    c1, c2 = get_indel_contexts(chrom, rpos+(len(ref)-len(alt)), radius, tb)
                except:
                    continue
                if break_point_type in ['del_start', 'all', 'del']:
                    if sample_breakpoints:
                        kmer_count[c2] += 1
                    else:
                        kmer_count[c2] += 1.0/(len(L)*4)
                if break_point_type in ['del_end', 'all', 'del']:
                    if sample_breakpoints:
                        kmer_count[c1] += 1
                    else:
                        kmer_count[c1] += 1.0/(len(L)*4)
    return kmer_count


def count_non_indels(tb, dreader, before, after, reverse_complement_method):
    kmer_count = defaultdict(int)
    chroms = tb.chroms()
    for chrom, pos, scale in dreader:
        if pos<before or pos > (chroms[chrom]-after):
            continue
        try:
            context = tb.sequence(chrom, pos-before, pos+after+1)
            if 'N' in context:
                continue
        except KeyError:
            print('Warning. Unknown position ignored:', chrom, pos, file=sys.stderr)
            continue
        if reverse_complement_method == "middle" and context[before] in ['T','G']:
            context = reverse_complement(context)
        elif reverse_complement_method == "lexicographic":
            context2 = reverse_complement(context)
            if context2 < context:
                context = context2
        elif reverse_complement_method == "both":
            context2 = reverse_complement(context)
            kmer_count[context2] += 1
        kmer_count[context] += 1
    return kmer_count


"""Module that contains the command line application."""

import gzip
import sys
import argparse
import py2bit
from kmer_counter.readers import *
from kmer_counter.counters import count_indels, count_non_indels

def main(args = None):
    """
    Run the main program.

    This function is executed when you type `kmer_counter` or `python -m kmer_counter`.

    Arguments:
        args: Arguments passed from the command line.

    Returns:
        An exit code.
    """
    parser = argparse.ArgumentParser(
        prog='kmer_counter',
        description='''
        Count k-mers at SNVs, indel breakpoints or in genomic regions
    ''')
    subparsers = parser.add_subparsers(dest='command', help='command. Must choose what kind of file you want to count.')
    parser.add_argument('-V', '--version', action='store_true')

    snv_parser = subparsers.add_parser('snv', 
        description='''Count k-mers at SNVs.
        The k-mer will be centered around the mutation base and if the reference base is G or T then the reverse_complement of the reference k-mer will be counted instead.
        '''
    )
    snv_parser.add_argument('ref_genome', help='Reference genome in 2bit format', type=str)
    snv_parser.add_argument('mutations', 
        type=argparse.FileType('r'), 
        help='A vcf-like file with SNVs. First four columns should be: Chrom, pos, ref, alt. '
        'Other columns are ignored. Non-SNV variants are ignored.')
    snv_parser.add_argument('-r', '--radius', type=int, metavar='R', default=1,
        help='Count the k-mer from R bases before a position to R bases '
             'after a position. For each position in the inputfile.')

    indel_parser = subparsers.add_parser('indel', description='Count k-mers at indels.')
    indel_parser.add_argument('ref_genome', help='Reference genome in 2bit format', type=str)
    indel_parser.add_argument('mutations',
        type=argparse.FileType('r'), 
        help='A sorted vcf-like file with indels. First four columns should be: Chrom, pos, ref, alt. '
        'Other columns are ignored. Non-indel variants are ignored. Indels should be left-aligned.')
    indel_parser.add_argument('type', choices=['ins', 'del_start', 'del_end', 'all', 'del'],
        help='What type of indel breakpoint do you want to count?')
    indel_parser.add_argument('-r', '--radius', default=1, type=int, 
        help='How many base pairs before indel_start_point or after indel_end_point should be included '
        'as context annotation.')
    indel_parser.add_argument('--sample', action="store_true", help='Randomly choose one of the possible positions instead of counting the expected (non integer) count for each possible position of an ambigously aligned indel.')
    #indel_parser.add_argument('-v', '--verbose', action='store_true')


    bg_parser = subparsers.add_parser('background', description='Count kmers in (regions of) a genome')
    bg_parser.add_argument('ref_genome',  type=str,
        help='Reference genome in 2bit format',)
    bg_parser.add_argument('--bed', type=str,
        help='bed-file describing regions that should be counted. May be gzipped.')
    #bg_parser.add_argument('--wig', type=str,
    #    help='wig-file describing regions that should be counted. May be gzipped. '
    #        'The context at a position will be weigthed by the value from the '
    #        'wig-file at that position. The output counts will thus be floats '
    #        'and not integers')
    bg_parser.add_argument('--all_autosomes', action="store_true",
        help='All parts of the autosomes will be counted')
    bg_parser.add_argument('-r', '--radius', type=int, metavar='R',
        help='Count the k-mer from R bases before a position to R bases '
             'after a position. For each position in the inputfile.')
    bg_parser.add_argument('--before_after', type=int, nargs=2, metavar=('X','Y'),
        help='count the k-mer from X bases before a position to Y bases after a position. '
        'For each position in the inputfile.')
    bg_parser.add_argument('--reverse_complement_method', type=str, choices=['none', 'middle', 'lexicographic', 'both'],
        help='"none" means that alle k-mers are counted unchanged. "middle" means that the reverse complement of a k-mer is counted if the middle position is not a "A" or "C". "lexicographic" means that the reverse_complement is counted if it has a smaller lexicographic order. Default is "middle" if --radius option is used and "lexicographic" if --before_after is used.')

    args = parser.parse_args(args)

    if args.version:
        from kmer_counter import __version__
        print("version:", __version__)
        print()
        return 0

    if args.command not in ['snv', 'indel', 'background']:
        print('Error: must specify command.')
        print()
        parser.print_help()
        return 0

    if 'ref_genome' not in args:
        print('Error: ref_genome (as 2bit file) must be specified')
        print()
        parser.print_help()
        return 0

    tb = py2bit.open(args.ref_genome)

    if args.command == 'indel':
        kmer_count = count_indels(args.mutations, tb, args.type, args.radius, args.sample)
    elif args.command == 'snv':
        dreader = PosReader(args.mutations, tb)
        kmer_count = count_non_indels(tb, dreader, args.radius, args.radius, 'middle')     
    elif args.command == 'background':
        if args.radius is None == args.before_after is None:
            raise Exception('Either the --radius or the --before_after option should be used (not both).')
        if not args.radius is None:
            assert args.radius>0
            before = args.radius
            after = args.radius
            if args.reverse_complement_method is None:
                args.reverse_complement_method = "middle"
        else:
            before, after = args.before_after
            assert before>=0
            assert after>=0
            if args.reverse_complement_method is None:
                args.reverse_complement_method = "none"        

        if not args.bed is None:
            if args.all_autosomes:
                raise Exception('Either --bed or--all_autosomes option should be used. Not both.')
            if args.bed.endswith('.bed.gz'):
                dreader = BedReader(gzip.open(args.bed, 'rt'))
            elif args.bed.endswith('.bed'):
                dreader = BedReader(open(args.bed))
            elif args.bed == '-':
                dreader = BedReader(sys.stdin)
            else:
                raise Exception('bed file should end with ".bed" or ".bed.gz"')
        elif args.all_autosomes:
            dreader = AllAutoReader(tb)
        else:
            raise Exception('Either --bed or--all_autosomes option should be used')
        kmer_count = count_non_indels(tb, dreader, before, after, args.reverse_complement_method)
    
    for x in kmer_count:
        print(x, kmer_count[x])
    
    return 0
